/*

				Javascript operations on buffers
				Roald Baudoux 2017
				Version : 2017-12-21

*/

// JS updates automatically to latest version whenever
// the file is edited in an external editor
autowatch = 1;

// post buffer size in console
function postsize(lebuffer) {
	if (lebuffer != null) {
		var unbuffer = new Buffer(lebuffer);
		post("Samples in "+lebuffer+ ": ", unbuffer.framecount());
	}
}

// application d'une enveloppe (durées libres en pourcentages)
// sur le son copié dans un autre buffer
// le buffer de destination est dimensionné en fonction
function envpercent(s,t, fadein, fadeout) {
	if (s != null && t != null) {
		// constraints time values
		fadein = Math.min(fadein, 100);
		fadein = Math.max(fadein, 0);
		fadeout = Math.min(fadeout, 100);
		fadeout = Math.max(fadeout, 0);
		if (fadein + fadeout > 100) {
			fadein = fadein - ((fadein + fadeout - 100) / 2);
			fadeout = fadeout - ((fadein + fadeout - 100) / 2);
			post("fade-in reduced to ", fadein, "% // fade-out reduced to ", fadeout, "%");
			post();
		}
		var source = new Buffer(s);
		var target = new Buffer(t);
		var amp, actuel;
		// dimensionnement du buffer d'arrivée
		target.send("sizeinsamps", source.framecount(), source.channelcount());
		// application de l'enveloppe et copie
		for (var i = 1; i < source.channelcount()+1; i++) {			
			for (var j = 0 ; j < source.framecount() ; j++) {	
				// a. attaque
				if (j < Math.floor(source.framecount() * fadein/100 )) {
					// calcul du facteur d'amplification
					amp = j/(fadein/100 * source.framecount());
					// application au son qui est dans le buffer
					// extraction de la valeur actuelle de l'échantillon
					actuel = source.peek(i, j, 1);
					// poke avec numéro de canal, position d'écriture et la ou les valeurs à écrire
					target.poke(i, j, actuel * amp);			
				}
				// b. sustain : copie du buffer d'entrée vers le buffer de sortie sans modification
				else if ((j >= Math.floor((fadein/100) * source.framecount())) && (j < Math.floor((1-fadeout/100) * source.framecount())))
				 {
					target.poke(i, j, source.peek(i, j, 1));
				}
				// c. release
				else {
					//post("release");
					//post();
					// calcul du facteur d'amplification
					amp = j - ((1-(fadeout/100)) * source.framecount()) ;
					amp = amp / (fadeout/100 * source.framecount());
					amp = 1 - amp;
					// affichage amp sur début et fin de fade-out
					//if (j == ((1-1/fadeout) * source.framecount()) + 1 || j == (source.framecount() - 1)) post ("amp = ", amp);
					// application au son qui est dans le buffer
					// extraction de la valeur actuelle de l'échantillon
					actuel = source.peek(i, j, 1);
					// poke avec numéro de canal, position d'écriture et la ou les valeurs à écrire
					target.poke(i, j, actuel * amp);
				}
			}
		}
	}
}

// application d'une enveloppe (durées libres en ms)
// sur le son copié dans un autre buffer
// le buffer de destination est dimensionné en fonction
// le taux d'échantillonnage du buffer doit être fourni
function envms(s,t, fadein, fadeout, sr) {
	if (s != null && t != null) {
		// constraints time values
		var source = new Buffer(s);
		var target = new Buffer(t);
		var amp, actuel;
		if (sr == null && sr <= 0) {
			post("samplerate not provided - 44 100 Hz taken");
			post();
		}
		// conversion des durées en échantillons
		fadein = fadein / 1000. * sr;
		fadein = Math.min(fadein, source.framecount());
		fadein = Math.max(fadein, 0);
		fadeout = fadeout / 1000. * sr;
		fadeout = Math.min(fadeout, source.framecount());
		fadeout = Math.max(fadeout, 0);
		if (fadein + fadeout > source.framecount()) {
			fadein = fadein - ((fadein + fadeout - source.framecount()) / 2);
			fadeout = fadeout - ((fadein + fadeout - source.framecount()) / 2);
			post("fade-in reduced to ", fadein, "s. // fade-out reduced to ", fadeout, "s.");
			post();
		}
		// dimensionnement du buffer d'arrivée
		target.send("sizeinsamps", source.framecount(), source.channelcount());
		// application de l'enveloppe et copie
		for (var i = 1; i < source.channelcount()+1; i++) {			
			for (var j = 0 ; j < source.framecount() ; j++) {	
				// a. attaque
				if (j < fadein) {
					// calcul du facteur d'amplification
					amp = j/fadein;
					// application au son qui est dans le buffer
					// extraction de la valeur actuelle de l'échantillon
					actuel = source.peek(i, j, 1);
					// poke avec numéro de canal, position d'écriture et la ou les valeurs à écrire
					target.poke(i, j, actuel * amp);			
				}
				// b. sustain : copie du buffer d'entrée vers le buffer de sortie sans modification
				else if ((j >= fadein) && (j < (source.framecount() - fadeout)))
				 {
					target.poke(i, j, source.peek(i, j, 1));
				}
				// c. release
				else {
					//post("release");
					//post();
					// calcul du facteur d'amplification
					amp = j - source.framecount() + fadeout ;
					amp = amp / fadeout;
					amp = 1 - amp;
					// affichage amp sur début et fin de fade-out
					//if (j == ((1-1/fadeout) * source.framecount()) + 1 || j == (source.framecount() - 1)) post ("amp = ", amp);
					// application au son qui est dans le buffer
					// extraction de la valeur actuelle de l'échantillon
					actuel = source.peek(i, j, 1);
					// poke avec numéro de canal, position d'écriture et la ou les valeurs à écrire
					target.poke(i, j, actuel * amp);
				}
			}
		}
	}
}

// application d'une enveloppe (durées libres en pourcentages)
// sur le son copié dans un autre buffer
// le buffer de destination est dimensionné en fonction
function envdelta(s,t) {
	if (s != null && t != null) {
		var source = new Buffer(s);
		var target = new Buffer(t);
		var amp, actuel;
		// dimensionnement du buffer d'arrivée
		target.send("sizeinsamps", source.framecount(), source.channelcount());
		// application de l'enveloppe et copie
		for (var i = 1; i < source.channelcount()+1; i++) {			
			for (var j = 0 ; j < source.framecount() ; j++) {	
				// a. attaque
				if (j < Math.floor(source.framecount() / 2 )) {
					// calcul du facteur d'amplification
					amp = 2 * j/source.framecount();
					// application au son qui est dans le buffer
					// extraction de la valeur actuelle de l'échantillon
					actuel = source.peek(i, j, 1);
					// poke avec numéro de canal, position d'écriture et la ou les valeurs à écrire
					target.poke(i, j, actuel * amp);			
				}
				// c. release
				else {
					//post("release");
					//post();
					// calcul du facteur d'amplification
					amp = j - (0.5 * source.framecount()) ;
					amp = 2 * amp / source.framecount();
					amp = 1 - amp;
					// affichage amp sur début et fin de fade-out
					//if (j == ((1-1/fadeout) * source.framecount()) + 1 || j == (source.framecount() - 1)) post ("amp = ", amp);
					// application au son qui est dans le buffer
					// extraction de la valeur actuelle de l'échantillon
					actuel = source.peek(i, j, 1);
					// poke avec numéro de canal, position d'écriture et la ou les valeurs à écrire
					target.poke(i, j, actuel * amp);
				}
			}
		}
	}
}

// boomerang
function boomerang(s,t) {
	if (s != null && t != null) {
		var source = new Buffer(s);
		var target = new Buffer(t);
		var amp, actuel;
		var samples = [];
		var samples_1 = [];
		var samples_2 = [];
		// dimensionnement du buffer d'arrivée
		target.send("sizeinsamps", source.framecount(), source.channelcount());
		post("taille du buffer: ", source.framecount());
		post();
		// traitement
		for (var i = 1; i < source.channelcount()+1; i++) {	
			samples = [];
			samples_1 = [];
			samples_2 = [];
			// copie de la première moitié du buffer dans le tableau 1
			samples_1 = source.peek(i, 0, Math.ceil(source.framecount() / 2));
			post("taille du tableau 1: ", samples_1.length);
			post();		
			// copie de la première moitié du buffer moins un échantillon
			// dans le tableau 2
			samples_2 = source.peek(i, 0, Math.floor(source.framecount() / 2));
			post("taille du tableau 2: ", samples_2.length);
			post();	
			// inversion du tableau 2
			samples_2.reverse();
			// concaténation des tableaux
			samples = samples_1.concat(samples_2);
			/*
			for (var j = 0; j < samples.length; j++) {
				samples[j] = Math.random() * 2  - 1;
			}
			*/
			post("taille du tableau global après concaténation: ", samples.length);
			post();		
			// copie du tableau à la fin du buffer cible
			target.poke(i, 0, samples);
		}
	}
}

// remplissage par du bruit blanc
function noise(lebuffer) {
	if (lebuffer != null) {
		var unbuffer = new Buffer(lebuffer);
		for (var i = 1; i < unbuffer.channelcount()+1; i++) {
			for (var j = 0 ; j < unbuffer.framecount() ; j++) {
				unbuffer.poke(i, j, Math.random() * 2 - 1);
			}
		}
		unbuffer.send("normalize", 1);
	}
}

// remplissage par une constante
function constant(lebuffer, myval) {
	if (lebuffer != null && myval != null) {
		var unbuffer = new Buffer(lebuffer);
		for (var i = 1; i < unbuffer.channelcount()+1; i++) {
			for (var j = 0 ; j < unbuffer.framecount() ; j++) {
				unbuffer.poke(i, j, myval);
			}
		}
	}
}

// cherche le maximum global et pour chaque canal
function postmax(lebuffer) {
	if (lebuffer != null) {
		var unbuffer = new Buffer(lebuffer);
		// tableau pour les maxima des canaux
		var indivmaxi = new Array(unbuffer.channelcount())
		// initialisation du tableau
		for (var i = 0; i < unbuffer.channelcount(); i++) indivmaxi[i] = -(Math.pow(2, 53) - 1);
		// maximal global
		var globalmaxi =-(Math.pow(2, 53) - 1);		// valeur de référence du Javascript en 64float
		// recherche des maxima
		for (var i = 1; i < unbuffer.channelcount()+1; i++) {
			for (var j = 0 ; j < unbuffer.framecount() ; j++) {
				if (unbuffer.peek(i, j, 1) > globalmaxi) globalmaxi = unbuffer.peek(i, j, 1);
				if (unbuffer.peek(i, j, 1) > indivmaxi[i-1]) indivmaxi[i-1] = unbuffer.peek(i, j, 1);
			}
		}
		// envoi des données en sortie
		// maximum global
		outlet(0, "global maximum: ", globalmaxi);
		// maximum canal par canal
		for (var i = 0; i < unbuffer.channelcount(); i++) {
			var channel = i + 1
			outlet(0, "ch. "+channel+ " maximum: ", indivmaxi[i]);
		}
	}
}

//clone un buffer sur un autre (redimensionné si nécessaire)
function copybuf(origin, dest) {
	if (origin != null && dest != null) {
		var myorigin = new Buffer(origin);
		var mydest = new Buffer(dest);
		// redimensionnement du buffer de destination à la taille du buffer d'origine
		mydest.send("sizeinsamps", myorigin.framecount(), myorigin.channelcount());
		for (var i = 1; i < myorigin.channelcount()+1; i++) {
			for (var j = 0; j < myorigin.framecount(); j++) {
				// copie d'un échantillon
				mydest.poke(i, j, myorigin.peek(i, j, 1));
			}
		}

	}
}

// applique une distorsion "x au carré multiplié par le signe de x"
function squared(lebuffer) {
	if (lebuffer != null) {
		var unbuffer = new Buffer(lebuffer);
		for (var i = 1; i < unbuffer.channelcount()+1; i++) {
			for (var j = 0; j < unbuffer.framecount(); j++) {
				// mise au carré de l'échantillon et multiplication par son signe pour éviter de
				// n'avoir que des échantillons positifs
				unbuffer.poke(i, j, Math.pow(unbuffer.peek(i, j, 1), 2) * ((unbuffer.peek(i, j, 1) >= 0) * 2 - 1));
			}
		}
	}
}

// application d'une distorsion "racine carrée de x multipliée par le signe de x"
function root2(lebuffer) {
	if (lebuffer != null) {
		var unbuffer = new Buffer(lebuffer);
		for (var i = 1; i < unbuffer.channelcount()+1; i++) {
			for (var j = 0; j < unbuffer.framecount(); j++) {
				// racine carrée de la valeur absolue de l'échantillon
				// et multiplication par son signe pour éviter
				// de n'avoir que des échantillons positifs
				var oldval = unbuffer.peek(i, j, 1);
				unbuffer.poke(i, j, Math.sqrt(Math.abs(oldval)) * (((oldval >= 0) * 2) - 1));
			}
		}
	}
}

// application d'une distorsion "racine énième de x multipliée par le signe de x"
function rootany(lebuffer, root) {
	if (lebuffer != null) {
		var unbuffer = new Buffer(lebuffer);
		root = Math.max(1, root);
		for (var i = 1; i < unbuffer.channelcount()+1; i++) {
			for (var j = 0; j < unbuffer.framecount(); j++) {
				// racine carrée de la valeur absolue de l'échantillon
				// et multiplication par son signe pour éviter
				// de n'avoir que des échantillons positifs
				var oldval = unbuffer.peek(i, j, 1);
				unbuffer.poke(i, j, Math.pow(Math.abs(oldval), 1/root) * (((oldval >= 0) * 2) - 1));
			}
		}
	}
}

// bégaiement aléatoire (ne fonctionne pas encore)
function stutter(origin, dest) {
	if (origin != null && dest != null) {
		var myorigin = new Buffer(origin);
		var mydest = new Buffer(dest);
		// redimensionnement du buffer de destination à la taille du buffer d'origine
		mydest.send("sizeinsamps", myorigin.framecount(), myorigin.channelcount());
		// effacement du buffer de destination
		mydest.send("clear");
		// définition d'un inhibiteur de copie tant qu'on est sur une section résultant d'une copie antérieure
		var inhibited = -1;
		for (var i = 0; i < myorigin.framecount(); i++) {
			// action possible uniquement si l'on est pas dans une partie où les échantillons résultent d'une copie
			if (i > inhibited) {
				// tirage aléatoire qui détermine si l'on bégaie ou non
				// on décide qu'on a 5% de chance d'avoir une répétition
				var chance = Math.floor((Math.random() * 100)) < 10;
				// si chance vaut 1, on a une répétition
				if (chance == 1) {
					post ("i = ", i);
					post();
					// taille à copier
					var numsamples = Math.floor((Math.random() * 1764) + 441); // de 10 à 50 ms à 44.1kHz en nombre d'échantillons
					// nombre de répétitions
					var numrepets = Math.floor(Math.random() * 11);
					var posdepartcopie = i;
					// met à jour l'inhibiteur
					inhibited = i + numsamples * numrepets;
					// copie et collage canal par canal
					// nombre de canaux
					for (var j = 0; j < myorigin.channelcount(); j++) {
						// copie dans un tableau échantillon par échantillon
						//var samples = new Array;
						//samples = myorigin.peek(j, posdepartcopie, numsamples);
						//var samples = {1, 0.9, 0.8, 0.7, 0.6, 0.5, 0.4, 0.3, 0.2, 0.1, 0., 0., 0.};
						// for (var k = 0; k < numsamples; k++) samples[k] = myorigin.peek(j, posdepartcopie+k, 1);
						// collage du tableau vers la destination
						// nombre de répétitions
						//for (var k = 1; k < numrepets+1; k++) {
							//échantillons
							//for (var l = 0; l < numsamples; l++) {
							//	mydest.poke(j, i + ( k * numsamples) + l, samples[l]);
							//mydest.poke(j, i + (k * numsamples), samples)
							//}
						//}
					}
					i = inhibited + 1;
				}
			}
			// simple copie sans répétition quand inhibition
			else {
				for (var j = 0; j < myorigin.channelcount(); j++) {
					mydest.poke(j, i, myorigin.peek(j,i,1));
				}
			}
		}
	}
	post ("stutter done!");
	post();
}

// distorsion par inversion de la valeur absolue des amplitudes par rapport à 1/-1.
// il est conseillé de normaliser le son préalablement
function antipodes(lebuffer) {
	if (lebuffer != null) {
		var unbuffer = new Buffer(lebuffer);
		for (var i = 1; i < unbuffer.channelcount()+1; i++) {
			for (var j = 0; j < unbuffer.framecount(); j++) {
				// transformation
				var oldamp = unbuffer.peek(i,j,1);
				var newamp = ((oldamp >= 0) * 2. - 1.) - oldamp ;
				unbuffer.poke(i, j, newamp);
			}
		}
	}
}

// normalisation des wavesets
function wavesetnorm(lebuffer) {
	if (lebuffer != null) {
		var unbuffer = new Buffer(lebuffer);
		// compteur de passages par 0 (il en faut 2)
		var crossings = 0;
		// variable de mémorisation de début du waveset
		var startpos = 0;
		// varible de mémorisation de fin du waveset
		var endpos = 0;
		// variable de mémorisation de maximum
		var localmax = 0;
		// normalisation des wavesets
		for (var i = 1; i < unbuffer.channelcount()+1; i++) {
			for (var j = 0; j < unbuffer.framecount(); j++) {
				var current = unbuffer.peek(i,j,1);
				var next = unbuffer.peek(i,j+1,1);
				// détection du maximum
				if (Math.abs(current) > localmax) localmax = Math.abs(current);
				// détection d'un franchissement de 0
				if ( (current >= 0) != (next >= 0.) ) {
					crossings++;
					// si deux franchissements, alors normalisation de la section
					if (crossings == 2) {
						endpos = j;
						// définition du facteur de normalisation
						var normfactor = 1/localmax;
						// normalisation
						for (var k = startpos; k < endpos; k++) {
							unbuffer.poke(i, k, unbuffer.peek(i,k,1) * normfactor);
						}
						// mise à jour des variables
						localmax = 0.;
						startpos = endpos + 1;
						crossings = 0;
					}
				}
			}
		}
	}
}

// remplacement des wavesets par des cycles sinusoïdaux
// le nombre de wavesets remplacés par un cycle est paramétré
function wavesetsine(lebuffer, numsets) {
	if (lebuffer != null && numsets != null) {
		var unbuffer = new Buffer(lebuffer);
		// nombre total de passages par zéros requis
		var totalsets = numsets * 2;
		// compteur de passages par 0 (il en faut 2 par waveset après le début)
		var crossings;
		// variable de mémorisation de début du waveset
		var startpos;
		// varible de mémorisation de fin du waveset
		var endpos;
		// remplacement des wavesets par des sinusoïdes
		for (var i = 1; i < unbuffer.channelcount()+1; i++) {
			crossings = 0;
			startpos = 0;
			endpos = 0;
			for (var j = 0; j < unbuffer.framecount(); j++) {
				var current = unbuffer.peek(i,j,1);
				var next = unbuffer.peek(i,j+1,1);
				// détection d'un franchissement de 0
				if ( (current >= 0) != (next >= 0.) ) {
					crossings++;
					// si n franchissements, alors on a un groupe de wavesets
					if (crossings == totalsets) {
						endpos = j;
						// remplacement par une sinusoïde
						for (var k = startpos; k < endpos; k++) {
							var amp = Math.sin(2 * Math.PI * (k - startpos)/(endpos - startpos));
							unbuffer.poke(i, k, amp);
						}
						// mise à jour des variables
						startpos = endpos + 1;
						crossings = 0;
					}
				}
			}
		}
	}
}

// mise à zéro d'un waveset sur deux
// ERREUR
function waveset1over2(lebuffer) {
	if (lebuffer != null) {
		var unbuffer = new Buffer(lebuffer);
		// compteur de passages par 0 (il en faut 2)
		var crossings;
		// variable de mémorisation de début du waveset
		var startpos;
		// varible de mémorisation de fin du waveset
		var endpos;
		// détecteur de parité du nombre de wavesets
		var wavesetnumberparity;
		// détection des wavesets et traitement
		for (var i = 1; i < unbuffer.channelcount()+1; i++) {
			crossings = 0;
			startpos = 0;
			endpos = 0;
			wavesetnumberparity = 0;
			for (var j = 0; j < unbuffer.framecount(); j++) {
				var current = unbuffer.peek(i,j,1);
				var next = unbuffer.peek(i,j+1,1);
				// détection d'un franchissement de 0
				if ( (current >= 0) != (next >= 0.) ) {
					crossings++;
					// si deux franchissements, waveset détecté
					if (crossings == 2) {
						endpos = j;
						switch (wavesetnumberparity) {
							case 0:		
								// old : ne fait rien
								// new : applique une enveloppe triangulaire
								// phase montante
								var amp;
								for (var k = startpos; k < Math.ceil(startpos+((endpos-startpos)/2)); k++) {
									amp = (k-startpos)/(Math.ceil(endpos-startpos)/2);
									unbuffer.poke(i, k, unbuffer.peek(i, k, 1) * amp);
								}
								// phase descendante
								for (var k = Math.ceil(startpos+((endpos-startpos)/2)); k < endpos; k++) {
									amp = 1 - (k-startpos)/(Math.ceil(endpos-startpos)/2);
									unbuffer.poke(i, k, unbuffer.peek(i, k, 1) * amp);
								}
								wavesetnumberparity = 1;
								break;
							case 1:
								// remplace le waveset par du silence
								for (var k = startpos; k < endpos; k++) {
									unbuffer.poke(i, k, 0.);
								}
								wavesetnumberparity = 0;
								break;
						}
						// mise à jour des variables
						startpos = endpos + 1;
						crossings = 0;
					}
				}
			}
		}
	}
}

// répartition alternative des wavesets d'un son mono
// sur un buffer (transformé d'office en buffer stereo)
function wavesetsplit2(origin, dest) {
	if (origin != null && dest != null) {
		var myorigin = new Buffer(origin);
		var mydest = new Buffer(dest);
		var samples_1 = [];
		var samples_2 = [];
		// copie du canal 1 du buffer d'origine sur deux tableaux
		samples_1 = myorigin.peek(1, 0, myorigin.framecount());
		samples_2 = myorigin.peek(1, 0, myorigin.framecount());
		// redimensionnement du buffer de destination à la taille du buffer d'origine
		// et sur deux canaux d'office
		mydest.send("sizeinsamps", myorigin.framecount(), 2);
		// effacement du buffer de destination
		mydest.send("clear");
		// compteur de passages par 0 (il en faut 2)
		var crossings = 0;
		// variable de mémorisation de début du waveset
		var startpos = 0;
		// variable de mémorisation de fin du waveset
		var endpos = 0;
		// détecteur de parité du nombre de wavesets
		var wavesetnumberparity = 0;
		// détection des wavesets et traitement
		for (var j = 0; j < samples_1.length; j++) {
			var current = samples_1[j];
			var next = samples_1[j+1];
			// détection d'un franchissement de 0
			if ( (current >= 0) != (next >= 0.) ) {
				crossings++;
				// si deux franchissements, waveset détecté
				if (crossings == 2) {
					endpos = j;
					switch (wavesetnumberparity) {
						case 0:		
							// maintient le canal gauche (tableau 1)
							// donc efface le tableau 2
							for (var k = startpos; k < endpos; k++) {
								samples_2[k] = 0.;
							}
							wavesetnumberparity = 1;
							break;
						case 1:
							// maintient le canal droite (tableau 2)
							// donc efface le tableau 1
							for (var k = startpos; k < endpos; k++) {
								samples_1[k] = 0.;
							}							
							wavesetnumberparity = 0;
							break;
					}
					// mise à jour des variables
					startpos = endpos + 1;
					crossings = 0;
				}
			}
			// ou le dernier waveset se termine d'office à la fin du buffer
			else if (j == (samples_1.length - 1)) {
				endpos = j;
				switch (wavesetnumberparity) {
					case 0:		
						// maintient le canal gauche (tableau 1)
						// donc efface le tableau 2 sur cette portion
						for (var k = startpos; k < endpos; k++) {
							samples_2[k] = 0.;
						}
						break;
					case 1:
						// maintient le canal droite (tableau 2)
						// donc efface le tableau 1 sur cette portion
						for (var k = startpos; k < endpos; k++) {
							samples_1[k] = 0.;
						}
						break;
				}
			}
		}
		// copie des tableaux vers le buffer de destination
		mydest.poke(1,0,samples_1);
		mydest.poke(2,0,samples_2);
	}
}

// remplacement des wavesets par des impulsions
function wavesetclick(lebuffer) {
	if (lebuffer != null) {
		var unbuffer = new Buffer(lebuffer);
		// compteur de passages par 0 (il en faut 2)
		var crossings = 0;
		// variable de mémorisation de début du waveset
		var startpos = 0;
		// variable de mémorisation de fin du waveset
		var endpos = 0;
		// traitement des wavesets
		for (var i = 1; i < unbuffer.channelcount()+1; i++) {
			for (var j = 0; j < unbuffer.framecount(); j++) {
				var current = unbuffer.peek(i,j,1);
				var next = unbuffer.peek(i,j+1,1);
				// détection d'un franchissement de 0
				if ( (current >= 0) != (next >= 0.) ) {
					crossings++;
					// si deux franchissements, waveset détecté
					if (crossings == 2) {
						endpos = j;
						// transformation
						// 1. extraction de l'amplitude maximale du waveset
						var mymax = 0;
						var mypos = 0;
						for (var k = startpos; k < endpos; k++) {
							// comparaison de la valeur absolue du maximum avec
							// la valeur absolue de l'échantillon courant
							if (Math.abs(unbuffer.peek(i, k, 1)) > Math.abs(mymax)) {
								mymax = unbuffer.peek(i, k, 1);
								mypos = k;
							}
						}
						// 2. Création de l'impulsion
						// et mise à zéro des autres échantillons
						for (var k = startpos; k < endpos; k++) {
							// hormis le maximum, tout le reste est mis à zéro
							if (k != mypos) {
								unbuffer.poke(i, k, 0.);
							}
						}
						// mise à jour des variables
						startpos = endpos + 1;
						crossings = 0;
					}
				}
			}
		}
	}
}

// duplication des wavesets
function wavesetduplicate(lebuffer, howmany) {
	if (lebuffer != null) {
		var unbuffer = new Buffer(lebuffer);
		howmany = Math.max(2, Math.floor(howmany));
		// tableau de mémoire globale pour tous les canaux
		var allsamples = [];
		post("length of allsamples: ", allsamples.length);
		post();
		// copie du buffer dans la mémoire globale canal par canal
		// on a un tableau de tableaux
		for (var i = 1; i < unbuffer.channelcount()+1; i++) {
			allsamples[i-1] = new Array();
			allsamples[i-1] = unbuffer.peek(i, 0, unbuffer.framecount());
			post("length of allsamples["+(i-1)+"]: ", allsamples[i-1].length);
			post();
		}
		//mémoire temporaire pour le canal en cours de traitement
		var samples = []; 	
		// compteur de passages par 0 (il en faut 2 pour arriver au waveset suivant)
		var crossings;
		// variables de mémorisation de début du waveset
		var startpos = [];
		// variables de mémorisation de fin du waveset
		var endpos = [];
		// position actuelle
		var currentstartpos;
		// détecteur de parité du nombre de wavesets
		var wavesetnumberparity;
		// compteur de wavesets
		var wavesetnumber;
		// traitement canal par canal
		for (var i = 1; i < unbuffer.channelcount()+1; i++) {
			samples = [];
			crossings = 0;
			startpos = [];
			endpos = [];
			currentstartpos = 0;
			wavesetnumberparity = 0;
			wavesetnumber = 0;
			//=======================================
			// détection des wavesets et mémorisation
			//=======================================
			for (var j = 0; j < unbuffer.framecount(); j++) {
				var current = unbuffer.peek(i,j,1);
				var next = unbuffer.peek(i,j+1,1);
				// détection d'un franchissement de 0
				if ( (current >= 0.) != (next >= 0.) ) {
					crossings++;
					// si deux franchissements, waveset détecté
					if (crossings == 2) {
						startpos[wavesetnumber] = currentstartpos;
						endpos[wavesetnumber] = j;
						// mise à jour des variables				
						currentstartpos = endpos[wavesetnumber] + 1;
						wavesetnumber++;
						crossings = 0;
					}
				}
				// last waveset stops at buffer's end
				else if (j == (unbuffer.framecount() - 1)) {
						startpos[wavesetnumber] = currentstartpos;
						endpos[wavesetnumber] = j;
						wavesetnumber++;
				}
			}
			post ("number of wavesets found: "+wavesetnumber);
			post();
			//=======================================
			// écriture des wavesets dans l'ordre
			//=======================================
			for (var j = 0; j < startpos.length; j++) {
				// copie chaque frameset autant de fois que demandé dans le
				// tableau du canal en cours de traitement
				var temparray = allsamples[i-1].slice(startpos[j], endpos[j]+1);
				for (var k = 0; k < howmany; k++) {
					samples = samples.concat(temparray);
				}				
			}
			// écriture du tableau de canal vers le tableau global
			//unbuffer.poke(i, 0, samples);
			for (var j = 0; j < samples.length; j++) {
				allsamples[i-1][j] = samples[j];
			}
		}
		// redimensionnement du buffer de destination
		unbuffer.send("sizeinsamps", allsamples[0].length, allsamples.length);
		// copie du tableau global vers le buffer, canal par canal
		for (var i = 1; i < unbuffer.channelcount()+1; i++) {
			unbuffer.poke(i, 0, allsamples[i-1]);
		}
	}
}

// brassage des wavesets (canal par canal)
function wavesetscramble(lebuffer) {
	if (lebuffer != null) {
		var unbuffer = new Buffer(lebuffer);
		var samples = []; 	//mémoire temporaire des échantillons
		// compteur de passages par 0 (il en faut 2 pour arriver au waveset suivant)
		var crossings;
		// variables de mémorisation de début du waveset
		var startpos = [];
		// variables de mémorisation de fin du waveset
		var endpos = [];
		// position actuelle
		var currentstartpos;
		// détecteur de parité du nombre de wavesets
		var wavesetnumberparity;
		// compteur de wavesets
		var wavesetnumber;
		// traitement canal par canal
		for (var i = 1; i < unbuffer.channelcount()+1; i++) {
			samples = [];
			crossings = 0;
			startpos = [];
			endpos = [];
			currentstartpos = 0;
			wavesetnumberparity = 0;
			wavesetnumber = 0;
			//=======================================
			// détection des wavesets et mémorisation
			//=======================================
			for (var j = 0; j < unbuffer.framecount(); j++) {
				var current = unbuffer.peek(i,j,1);
				var next = unbuffer.peek(i,j+1,1);
				// détection d'un franchissement de 0
				if ( (current >= 0.) != (next >= 0.) ) {
					crossings++;
					// si deux franchissements, waveset détecté
					if (crossings == 2) {
						startpos[wavesetnumber] = currentstartpos;
						endpos[wavesetnumber] = j;
						// affichage
						post("startpos: ", currentstartpos);
						post(" // endpos: ", j);
						post(" // size: ", j - currentstartpos);
						post();
						// mise à jour des variables				
						currentstartpos = endpos[wavesetnumber] + 1;
						wavesetnumber++;
						crossings = 0;
					}
				}
				// last waveset stops at buffer's end
				else if (j == (unbuffer.framecount() - 1)) {
						startpos[wavesetnumber] = currentstartpos;
						endpos[wavesetnumber] = j;
						// affichage
						post("startpos: ", currentstartpos);
						post(" // endpos: ", j);
						post(" // size: ", j - currentstartpos);
						post();
				}
			}
			post ("wavesetscramble  - number of wavesets found: "+wavesetnumber);
			post();
			//=======================================
			// écriture des wavesets dans le désordre
			//=======================================
			// création d'un tableau représentant les index
			var mysets = [];
			// remplissage par des nombres représentant des positions
			// des tableaux contenant les positions de début et fin des wavesets
			for (var j = 0; j < wavesetnumber; j++) mysets[j] = j;
			// copie des wavesets dans le désordre dans le tableau provisoire
			var currentpos = 0;
			for (var j = 0; j < unbuffer.framecount(); j++) {
				// tirage aléatoire d'un index dans les wavesets
				var tirage = Math.floor(Math.random() * mysets.length);
				// copie du waveset dans le buffer temporaire
				var temparray = [];
				temparray = unbuffer.peek(i, startpos[mysets[tirage]], 
					endpos[mysets[tirage]]-startpos[mysets[tirage]]+1);
				samples = samples.concat(temparray);
				// déplacement dans le buffer temporaire
				currentpos = currentpos + (endpos[mysets[tirage]]-startpos[mysets[tirage]]);
				// élimination de la position tirée
				mysets.splice(tirage, 1);
				if (mysets.length == 0) break;
			}
			// écriture du tableau temporaire vers le buffer
			unbuffer.poke(i, 0, samples);
		}
	}
}

// brassage des wavesets par paquets de 50 (canal par canal)
function waveset50scramble(lebuffer) {
	if (lebuffer != null) {
		var unbuffer = new Buffer(lebuffer);
		var samples = new Array(0); 	//mémoire temporaire des échantillons
		// compteur de passages par 0 (il en faut 100)
		var crossings;
		// variables de mémorisation de début du waveset
		var startpos = new Array();
		// variables de mémorisation de fin du waveset
		var endpos = new Array();
		// position actuelle
		var currentstartpos;
		// détecteur de parité du nombre de wavesets
		var wavesetnumberparity;
		// compteur de wavesets
		var wavesetnumber;
		//=======================================
		// détection des wavesets et mémorisation
		//=======================================
		// traitement canal par canal
		for (var i = 1; i < unbuffer.channelcount() + 1; i++) {
			samples = [];
			crossings = 0;
			startpos = [];
			endpos = [];
			currentstartpos = 0;
			wavesetnumberparity = 0;
			wavesetnumber = 0;
			for (var j = 0; j < unbuffer.framecount(); j++) {
				var current = unbuffer.peek(i,j,1);
				var next = unbuffer.peek(i,j+1,1);
				// détection d'un franchissement de 0
				if ( (current >= 0.) != (next >= 0.) ) {
					crossings++;
					// si 100 franchissements, 50 wavesets détectés
					if (crossings == 100) {
						startpos[wavesetnumber] = currentstartpos;
						endpos[wavesetnumber] = j;
						post("group of 50 waveset's duration:", endpos[wavesetnumber]-startpos[wavesetnumber]);
						post();
						// mise à jour des variables				
						currentstartpos = endpos[wavesetnumber] + 1;
						wavesetnumber++;
						crossings = 0;
					}
				}
				// last waveset stops at buffer's end
				else if (j == (unbuffer.framecount() - 1)) {
					startpos[wavesetnumber] = currentstartpos;
					endpos[wavesetnumber] = j;
					// affichage
					post("startpos: ", currentstartpos);
					post(" // endpos: ", j);
					post(" // size: ", j - currentstartpos);
					post();
				}
			}
			post ("wavesetscramble  - number of wavesets found: "+wavesetnumber);
			post();
			//=======================================
			// écriture des wavesets dans le désordre
			//=======================================
			// création d'un tableau représentant les index
			var mysets = new Array(wavesetnumber);
			// remplissage par des nombres représentant des positions
			// des tableaux contenant les positions de début et fin des wavesets
			for (var j = 0; j < wavesetnumber; j++) mysets[j] = j;
			// copie des wavesets dans le désordre dans le tableau provisoire
			var currentpos = 0;
			for (var j = 0; j < unbuffer.framecount(); j++) {
				// tirage aléatoire d'un index dans les wavesets
				var tirage = Math.floor(Math.random() * mysets.length);
				// copie du waveset dans le buffer temporaire
				var temparray = new Array();
				temparray = unbuffer.peek(i, startpos[mysets[tirage]], 
					endpos[mysets[tirage]]-startpos[mysets[tirage]]+1);
				samples = samples.concat(temparray);
				// déplacement dans le buffer temporaire
				currentpos = currentpos + (endpos[mysets[tirage]]-startpos[mysets[tirage]]);
				// élimination de la position tirée
				mysets.splice(tirage, 1);
				if (mysets.length == 0) break;
			}
			// écriture du buffer temporaire vers le buffer final
			for(var j = 0; j < samples.length; j++) {
				unbuffer.poke(i, j, samples[j]);
				// ajout d'un fade-in de 20 échantillons
				if (j < 300) unbuffer.poke(i, j, unbuffer.peek(i, j, 1) * j/300);
				// ajout d'un fade-out de 20 échantillons
				if (j > samples.length - 300)
					unbuffer.poke(i, j, unbuffer.peek(i, j, 1) * (1-((j - samples.length + 300)/300)));
			}
		}
	}
}

// multiplie la forme d'onde par du bruit
function mynoise(lebuffer) {
	if (lebuffer != null) {
		var unbuffer = new Buffer(lebuffer);
		for (var i = 1; i < unbuffer.channelcount()+1; i++) {
			for (var j = 0; j < unbuffer.framecount(); j++) {
				current = unbuffer.peek(i, j, 1);
				unbuffer.poke(i, j, current * Math.random());
			}
		}
	}
}

// calcule le signal différentiel ou dérivé (delta)
function delta(lebuffer) {
	if (lebuffer != null) {
		var unbuffer = new Buffer(lebuffer);
		var previous;
		for (var i = 1; i < unbuffer.channelcount()+1; i++) {
			previous = 0;
			for (var j = 0; j < unbuffer.framecount(); j++) {
				var current = unbuffer.peek(i, j, 1);
				current = current - previous;		
				unbuffer.poke(i, j, current);
				previous = current;
			}
		}
	}
}

//transformation d'un buffer mono en stéréo et duplication du son
function monotostereo(lebuffer) {
	if (lebuffer != null) {
		var unbuffer = new Buffer(lebuffer);
		var length = unbuffer.framecount();
		var samples = new Array;
		samples = unbuffer.peek(1, 0, length);
		// redimensionnement du buffer sur deux canaux
		unbuffer.send("sizeinsamps", unbuffer.framecount(), 2);
		// copie des échantillons sur le buffer redimensionné
		unbuffer.poke(1, 0, samples);
		unbuffer.poke(2, 0, samples);
	}
}

//transformation d'un buffer mono en multicanal (et duplication)
function monotoany(lebuffer, numchans) {
	if (lebuffer != null) {
		var unbuffer = new Buffer(lebuffer);
		var length = unbuffer.framecount();
		var samples = new Array;
		samples = unbuffer.peek(1, 0, length);
		// redimensionnement du buffer sur deux canaux
		unbuffer.send("sizeinsamps", unbuffer.framecount(), numchans);
		// copie des échantillons sur le buffer redimensionné
		for (var i = 1; i < numchans+1; i++) {
			unbuffer.poke(i, 0, samples);
		}
	}
}

// sommation d'un buffer multicanal en mono
// et division de l'amplitude par le nombre de canaux
function anytomono(lebuffer, numchans) {
	if (lebuffer != null) {
		var unbuffer = new Buffer(lebuffer);
		var samples = new Array(unbuffer.channelcount());
		// insertion d'un tableau dans le premier sous-tableau et copie du premier canal
		samples[0] = new Array();
		samples[0] = unbuffer.peek(1, 0, unbuffer.framecount());
		// copie des autres canaux dans les sous-tableaux
		for (var i = 1; i < unbuffer.channelcount()+1; i++) {
			samples[i] = new Array();
			samples[i] = unbuffer.peek(i, 0, unbuffer.framecount());
			post ("size of this channel: ", samples[i].length);
			post();
		}
		// mixage et adaptation de l'amplitude
		for (var i = 1; i < unbuffer.channelcount()+1; i++) {
			for (var j = 1; j < unbuffer.framecount(); j++) {
				samples[0][j] = samples[0][j] + samples[i][j];
				// changes amplitude when all channels have been mixed
				if (i == unbuffer.channelcount()) {
					samples[0][j] = samples[0][j] / unbuffer.channelcount();
				}
			}
		}
		// redimensionnement du buffer sur un canal
		unbuffer.send("sizeinsamps", unbuffer.framecount(), 1);
		// copie des échantillons sur le buffer redimensionné
		unbuffer.poke(i, 0, samples[0]);
	}
}

// création d'un buffer stéréo à partir de deux buffers mono
function join2(o1, o2, d) {
	if (01 != null && 02 != null && d != null) {
		var origin_1 = new Buffer(o1);
		var origin_2 = new Buffer(o2);
		var dest = new Buffer(d);
		var samples_1 = [];
		var samples_2 = [];
		// copie dans les tableaux
		samples_1 = origin_1.peek(1, 0, origin_1.framecount());
		samples_2 = origin_2.peek(1, 0, origin_2.framecount());
		// redimensionnement du buffer d'arrivée
		dest.send("sizeinsamps", Math.max(samples_1.length, samples_2.length), 2);
		// copie des échantillons sur le buffer redimensionné
		dest.poke(1, 0, samples_1);
		dest.poke(2, 0, samples_2);
	}
}

// inversion temporelle
function reverse(lebuffer){
	if (lebuffer != null) {
		var unbuffer = new Buffer(lebuffer);
		var samples = new Array();
		for (var i = 1; i < unbuffer.channelcount()+1; i++) {
			samples = unbuffer.peek(i, 0, unbuffer.framecount());
			samples.reverse();
			unbuffer.poke(i, 0, samples);
		}
	}
}

// inversion de polarité
function invert(lebuffer){
	if (lebuffer != null) {
		var unbuffer = new Buffer(lebuffer);
		var samples = new Array();
		for (var i = 1; i < unbuffer.channelcount()+1; i++) {
			samples = unbuffer.peek(i, 0, unbuffer.framecount());
			for (var j = 0; j < unbuffer.framecount(); j++) {
				samples[j] = -samples[j];
			}
			unbuffer.poke(i, 0, samples);
		}
	}
}

//mixage de deux buffers (amplitudes linéaires)
function mix(o1, amp_1, o2, amp_2, t) {
	if (o1 != null && o2 != null && t != null) {
		var origin_1 = new Buffer(o1);
		var origin_2 = new Buffer(o2);
		var target = new Buffer(t);
		var samples_1 = new Array();
		var samples_2 = new Array();
		var samples_t = new Array();
		// check size and channel number or each buffer
		// should check if target is also one of origins
		var channelcount_t = Math.max(origin_1.channelcount(), origin_2.channelcount());
		var framecount_t = Math.max(origin_1.framecount(), origin_2.framecount());
		// target is resized according to requirements from origins
		target.send("sizeinsamps", framecount_t, channelcount_t);
		// now we can start to mix channel by channel
		for (var i = 1; i < channelcount_t+1; i++) {
			// first we copy the source buffers into arrays
			samples_1 = origin_1.peek(i, 0, origin_1.framecount());
			samples_2 = origin_2.peek(i, 0, origin_2.framecount());
			// then we add zeros to shortest array f required
			if (origin_1.framecount() != origin_2.framecount()) {
				if (origin_1.framecount() < origin_2.framecount()) {
					for (var j = origin_1.framecount(); j < origin_2.framecount(); j++) {
						samples_1[j] = 0.;
					}
				}
				else {
					for (var j = origin_2.framecount(); j < origin_1.framecount(); j++) {
						samples_2[j] = 0.;
					}
				}
			}
			// and we mix into another array
			for (var j = 0; j < framecount_t; j++) {
				samples_t[j] = (samples_1[j] * amp_1) + (samples_2[j] * amp_2);
			}
			// finally we copy the target array into the target buffer
			target.poke(i, 0, samples_t);
		}
	}
}

//mixage de deux buffers (amplitudes en dB)
function mixdb(o1, dB_1, o2, dB_2, t) {
	if (o1 != null && o2 != null && t != null) {
		var origin_1 = new Buffer(o1);
		var origin_2 = new Buffer(o2);
		var target = new Buffer(t);
		var samples_1 = new Array();
		var samples_2 = new Array();
		var samples_t = new Array();
		// converts dBs to linear amplitudes
		var amp_1 = Math.pow(10, dB_1/20.);
		var amp_2 = Math.pow(10, dB_2/20.);
		// check size and channel number or each buffer
		// should check if target is also one of origins
		var channelcount_t = Math.max(origin_1.channelcount(), origin_2.channelcount());
		var framecount_t = Math.max(origin_1.framecount(), origin_2.framecount());
		// target is resized according to requirements from origins
		target.send("sizeinsamps", framecount_t, channelcount_t);
		// now we can start to mix channel by channel
		for (var i = 1; i < channelcount_t+1; i++) {
			// first we copy the source buffers into arrays
			samples_1 = origin_1.peek(i, 0, origin_1.framecount());
			samples_2 = origin_2.peek(i, 0, origin_2.framecount());
			// then we add zeros to shortest array f required
			if (origin_1.framecount() != origin_2.framecount()) {
				if (origin_1.framecount() < origin_2.framecount()) {
					for (var j = origin_1.framecount(); j < origin_2.framecount(); j++) {
						samples_1[j] = 0.;
					}
				}
				else {
					for (var j = origin_2.framecount(); j < origin_1.framecount(); j++) {
						samples_2[j] = 0.;
					}
				}
			}
			// and we mix into another array
			for (var j = 0; j < framecount_t; j++) {
				samples_t[j] = (samples_1[j] * amp_1) + (samples_2[j] * amp_2);
			}
			// finally we copy the target array into the target buffer
			target.poke(i, 0, samples_t);
		}
	}
}

// lowpass filter by averaging over 2 samples
function lowpass2(s, t)
{
    var source = new Buffer(s);
	var target = new Buffer(t);
	var input = new Array();
	var output = new Array();
    // target is resized according to requirements from source
	target.send("sizeinsamps", source.framecount(), source.channelcount());
	// filtering by most simple lowpass filter
	// see : https://ccrma.stanford.edu/~jos/fp/Definition_Simplest_Low_Pass.html
	for(var i = 1; i < source.channelcount()+1; i++) {
		// copy source buffer in input array
		input = source.peek(i, 0, source.framecount());
		// filter
		output[0] = input[0];
		for(var j = 1; j < source.framecount(); j++) { 
			output[j] = (input[j]+input[j-1])/2;
    	}
    	// copy output array into target buffer
    	target.poke(i, 0, output);
	}  
}

// lowpass filter by averaging over 100 samples
function lowpass100(s, t)
{
    var source = new Buffer(s);
	var target = new Buffer(t);
	var input = new Array();
	var output = new Array();
	// tableau employé pour le filtrage
	var running = new Array();
    // target is resized according to requirements from source
	target.send("sizeinsamps", source.framecount(), source.channelcount());
	// filtering
	for(var i = 1; i < source.channelcount()+1; i++) {
		// copy source buffer in input array
		input = source.peek(i, 0, source.framecount());
		// fill beginning of output with input
		output = source.peek(i, 0, 99);
		// fill running array with zeros but last position
		running = source.peek(i, 0, 99);
		// actual filtering
		for(var j = 99; j < source.framecount(); j++) { 
			// on ajoute un échantillon à la fin du tableau
			running.splice(99,1,input[j]);
			// on fait la moyenne et on l'écrit dans la sortie
			output[j] = running.reduce(arraysum)/100;
			// on enlève un échantillon au début du tableau
			running.splice(0,1);
    	}
    	// copy output array into target buffer
    	target.poke(i, 0, output);
	}  
}    

// =================================================================
//						AUXILIARY FUNCTIONS
// =================================================================

// auxiliary function to get the sum of all values in an array
function arraysum(total, num) {
    return total + num;
}